/* **************************************************************************
 * @version $Id: oglBMFont.c,v 0.2 2006/04/03 12:03:00 legolas558 Exp $
 *
 * File:		oglBMFont.c
 * Content:		main source code for the bm_font object and related
 *				methods/members
 * Notes:		uses GLFW framework compiled into a static library
 *				uses RWops object for I/O
 *
 * Copyright(c) 2006 by legolas558
 *
 * https://sourceforge.net/projects/oglbmfont
 *
 * This software is released under a BSD license. See LICENSE.TXT
 * You must accept the license before using this software.
 *
 * **************************************************************************/

#include "oglBMfont.h"

#include "glfw.h"

#define		GL_Y_FIX(f)			(1 - (f))		// macro used to apply the y-inversion fix

// compile all charset characters into display lists
void	bmf__compile(bm_font *font)
{
	int chr;

	font->base_list=glGenLists(BMF_MAX_ASCII);
	glBindTexture(GL_TEXTURE_2D, font->fontTex);

	for (chr=0; chr<BMF_MAX_ASCII; chr++)
	{
		glNewList(font->base_list+chr,GL_COMPILE);

		if (font->chars[chr].w) {
			glBegin(GL_QUADS);
	glTexCoord2f(	(float) font->chars[chr].x / (float) font->w,
					(float) GL_Y_FIX((font->chars[chr].y+font->chars[chr].h) / (float) font->h));
	glVertex2d(		(float) font->chars[chr].x_ofs,
					(float) font->chars[chr].h + font->chars[chr].y_ofs);

	glTexCoord2f(	(float) (font->chars[chr].x + font->chars[chr].w) / (float) font->w,
					(float) GL_Y_FIX((font->chars[chr].y + font->chars[chr].h) / (float) font->h));
	glVertex2d(		(float) font->chars[chr].w + font->chars[chr].x_ofs,
					(float) font->chars[chr].h + font->chars[chr].y_ofs);

	glTexCoord2f(	(float) (font->chars[chr].x+font->chars[chr].w) / (float) font->w,
					(float) GL_Y_FIX( font->chars[chr].y / (float) font->h));
	glVertex2d(		(float) font->chars[chr].w + font->chars[chr].x_ofs,
					(float) font->chars[chr].y_ofs);

	glTexCoord2f(	(float) font->chars[chr].x / (float) font->w,
					(float) GL_Y_FIX( font->chars[chr].y / (float) font->h));
	glVertex2d(		(float) font->chars[chr].x_ofs,
					(float) font->chars[chr].y_ofs);
			glEnd();										
			glTranslatef((float)(font->chars[chr].x_advance-font->chars[chr].x_ofs), 0,0);							
		} else // if char has no width, treat it like a space
			glTranslatef((float)font->base, 0, 0);
		glEndList();	
	}				
}

// main function used to load a BMF font
// note: must be called within a valid OpenGL context because it allocates
// a texture and creates display lists for the font
bool	bmf_load(bm_font *font, char *bmf_path) {
	RWops	*bmf;
	int		i;
	long	p;
	char	tga_path[BMF_MAX_PATH];
	char	spaces[100];
	BMchar	temp_char;

	bmf = RWFromFile(bmf_path, "r");				// could not open file for read
	if (!bmf) return false;

	RWskipline(bmf);								// skip the first info line

	// read some basic attributes
	if (RWsscanf(bmf, "common lineHeight=%d base=%d scaleW=%d scaleH=%d pages=%d\n",
		&font->line_h, &font->base, &font->w, &font->h, &font->pages) != 5) {
			RWclose(bmf);
			return false;
	}

	// allocate (and init to 0) the chars (up to BMF_MAX_ASCII=256)
	font->chars = (BMchar *) zcalloc(BMF_MAX_ASCII, sizeof(BMchar));
NextChar:
	// read a character from the input file
	// note that the ordinal 'i' is also read and in a second pass temp_char is
	// swapped into the main array of chars
	if (RWsscanf(bmf, "char id=%d%[ ]x=%d%[ ]y=%d%[ ]width=%d%[ ]height=%d%[ ]xoffset=%d%[ ]yoffset=%d%[ ]xadvance=%d%[ ]page=%d%[ ]\n",
			&i, spaces,
			&temp_char.x, spaces,
			&temp_char.y, spaces,
			&temp_char.w, spaces,
			&temp_char.h, spaces,
			&temp_char.x_ofs, spaces,
			&temp_char.y_ofs, spaces,
			&temp_char.x_advance, spaces,
			&temp_char.page, spaces) == 9*2)
		{
			font->chars[i] = temp_char;
			goto NextChar;
		}

	RWclose(bmf);										// close the source file

	p = (long)strrchr(bmf_path, '.');
	if (!p) goto Fail;
	p-=(long)bmf_path;
	memcpy(tga_path, bmf_path, p+1);
	strcpy(tga_path+p+1, "tga");
	
	glGenTextures(1, &font->fontTex);					// create the main font texture
	glBindTexture(GL_TEXTURE_2D, font->fontTex);		// and set as current
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	if (!glfwLoadTexture2D(tga_path, 0)) {				// try to load the font TGA file
		glDeleteTextures(1, &font->fontTex);
Fail:
		zfree(font->chars);
		return false;
	}

	font->scale = 1.f;
	bmf__compile(font);									// (see above) compile the charset
	return true;										// into an OpenGL display list
}

// calculate the approx. width of a string of text
// note: no kerning info is currently evaluated
int		bmf__width(bm_font *font, char *text) {
	int	w, l, i;
	w = 0;
	l = strlen(text);
	for (i=0;i < l; i++) {
		w += font->chars[text[i]].w;
	}
	return w+l;
}

// destroy the charset
void	bmf_destroy(bm_font *font) {
	glDeleteLists(font->base_list, BMF_MAX_ASCII);
	glDeleteTextures(1, &font->fontTex);
	zfree(font->chars);
//	zfree(font->kerninfo);
}

void	ogl___font_in(ogl_font *font);

void	bmf___font_in(ogl_font *font) {
	ogl___font_in(font);
	glBindTexture(GL_TEXTURE_2D, font->fontTex);
}

#include "oglfont.c"

